
//
//   Many functions in this file adapted from 
//                   festival/src/modules/base/word.cc
//
/*************************************************************************/
/*                                                                       */
/*                Centre for Speech Technology Research                  */
/*                     University of Edinburgh, UK                       */
/*                       Copyright (c) 1996,1997                         */
/*                        All Rights Reserved.                           */
/*                                                                       */
/*  Permission is hereby granted, free of charge, to use and distribute  */
/*  this software and its documentation without restriction, including   */
/*  without limitation the rights to use, copy, modify, merge, publish,  */
/*  distribute, sublicense, and/or sell copies of this work, and to      */
/*  permit persons to whom this work is furnished to do so, subject to   */
/*  the following conditions:                                            */
/*   1. The code must retain the above copyright notice, this list of    */
/*      conditions and the following disclaimer.                         */
/*   2. Any modifications must be clearly marked as such.                */
/*   3. Original authors' names are not deleted.                         */
/*   4. The authors' names are not used to endorse or promote products   */
/*      derived from this software without specific prior written        */
/*      permission.                                                      */
/*                                                                       */
/*  THE UNIVERSITY OF EDINBURGH AND THE CONTRIBUTORS TO THIS WORK        */
/*  DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING      */
/*  ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT   */
/*  SHALL THE UNIVERSITY OF EDINBURGH NOR THE CONTRIBUTORS BE LIABLE     */
/*  FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES    */
/*  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN   */
/*  AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,          */
/*  ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF       */
/*  THIS SOFTWARE.                                                       */
/*                                                                       */
/*************************************************************************/

#include <stdlib.h>
#include "festival.h"
#include "lexicon.h"

static void check_multisyllabic(LISP entry);
static EST_Item *add_syllable(EST_Utterance *u,  LISP syl);
static EST_Item *add_segment(EST_Utterance *u, const EST_String &s);
static LISP specified_word_pronunciation(EST_Item *w, LISP lpos);


LISP FL_Lyricize_Utt(LISP utt){

  cout << "* FL_Lyricize\n";

  // Look up words in lexicon or use letter-to-sound rules 
  //  and create syllable and segment relations
  EST_Utterance *u = get_c_utt(utt);
  EST_Item *lyr, *lyr_tok,*syl,*seg,*note;
  LISP entry,s,p,lpos;
  EST_String pos;

  // add silence at the beginning of the utt
  add_segment(u, ph_silence());

  for (lyr=u->relation("Word")->first(); lyr != 0; lyr = next(lyr))  {

    // we might actually want homograph disambiguation at some point,
    //  or a convention allowing users to specify POS,
    //  but otherwise, we don't need POS yet....
    lpos = NIL;
    pos = ffeature(lyr,"hg_pos"); // explicit homograph pos disambiguation
    if (pos == "0")
      pos = ffeature(lyr,"pos");
    if (pos != "0")
      lpos = rintern(pos);

    //  Check if there is an explicitly given pronunciation attached
    //   to the Word before going to the lexicon.  
    //    --  also use this later for the "syllabified orthography" case
    cout << "try to pronounce \"" << lyr->name() << "\""<< endl;
    if ((entry = specified_word_pronunciation(lyr,lpos)) == NIL)
      entry = lex_lookup_word(lyr->name(),lpos);

    // give a POS if we have one
    if (lpos == NIL)
      lyr->set("pos",get_c_string(car(cdr(entry))));


    // Haven't made it take care of multisyllable LyricTokens yet - 
    //   we may get in trouble later, so do this check for now.....
    check_multisyllabic(entry);


    // for each syllable
    for (s=car(cdr(cdr(entry))); s != NIL; s=cdr(s)){
      
      syl = add_syllable(u, s);
      
      // append syl as daughter to Word in SylStructure
      append_daughter(lyr,"SylStructure",syl);

      // append syl as daughter to Word in NoteStruct
      append_daughter(lyr,"NoteStruct",syl);

      // append segments as daughters of Syllable 
      //    in SylStructure only
      for (p=car(car(s)); p != NIL; p=cdr(p)){
	seg = add_segment(u,get_c_string(car(p)));
	append_daughter(syl,"SylStructure",seg);
      }

      // find notes that are associated with this syllable
      //  and append as daughters in NoteStruct only
      lyr_tok = parent(parent(syl->as_relation("SylStructure")));
      note = daughter1(lyr_tok->as_relation("NoteWordAssoc"));
      for ( ; note != NULL; note=next_sibling(note)){
	append_daughter(syl,"NoteStruct",note);
      }
    }    
  }

  // add silence at the end of the utt
  add_segment(u, ph_silence());
  
  return utt;
}
static void check_multisyllabic(LISP entry){
  int k = 0;
  LISP s,p;
  
  for (s=car(cdr(cdr(entry))); s != NIL; s=cdr(s))
    k++;

  if (k>1){
    cerr << "Can't have multisyllable word on one note:  (";
    for (s=car(cdr(cdr(entry))); s != NIL; s=cdr(s)){
      cerr << "(";
      for (p=car(car(s)); p != NIL; p=cdr(p)){
	cerr << get_c_string(car(p)) << " ";
      }
      cerr << ")";
    }
    cerr << ")";
    EST_error("...");
  }
}
static EST_Item *add_syllable(EST_Utterance *u,  LISP entry){
  EST_Item *item = u->relation("Syllable")->append();
  int stress = get_c_int(car(cdr(car(entry))));

  // make syl name for debug
  LISP p = car(car(entry));
  EST_String name = get_c_string(car(p));  
  for (p=cdr(p); p != NIL; p=cdr(p)) 
    name = name + "_" + get_c_string(car(p));
  
  item->set_name(name); 
  item->set("stress",stress);
  
  return item;
}

static EST_Item *add_segment(EST_Utterance *u, const EST_String &s){
  EST_Item *item = u->relation("Segment")->append();   
  item->set_name(s);
  return item;
}

static LISP specified_word_pronunciation(EST_Item *w, LISP lpos){
  //  If there is a phoneme feature on w or the LyricToken related to 
  //  w use that as the pronunciation.  Note the value will be a string
  //  from which a list can be read.
  EST_String p;
  
    if (((p = ffeature(w,"phonemes")) != "0") ||
	((p = ffeature(w,"R:SylStructure.parent.phonemes")) != "0")){
    LISP phones = read_from_lstring(strintern(p));
    
    return cons(strintern(w->name()),
		cons(lpos,
		     cons(lex_syllabify(phones),NIL)));
  }
  else
    return NIL;
  
}
