#include <stdlib.h>
#include "festival.h"
#include "MIDI/Song.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "fl_midi_parse.h"
#include "fl_midi_print.h"
#include "fl_midi_ext.h"
#include "fl_io.h"

static float BaseFreq[12] = {8.17575, 8.662, 9.177, 
			     9.72275, 10.30075, 10.9135, 
			     11.56225, 12.24975, 12.97825, 
			     13.75, 14.5675, 15.43375};

static EST_String NoteName[12] = {"C", "C#", "D", "D#", "E", "F", 
				  "F#","G", "G#", "A", "A#", "B"};

float MIDIticks_to_sec(int ticks);
void MIDIpitch_to_freq_and_name(int note_num, float &freq, EST_String &name);
static EST_Utterance *get_current_utt(void);
static void set_current_utt(EST_Utterance *u);
EST_Utterance *Utt;  // this avoids including all the EST stuff in the MIDI lib
 
void update_tempo(unsigned long time, int bpm);
static float PrevSecAnchor = 0.0;
static unsigned long PrevTicksAnchor = 0;
static short CurrentTempo = 120;  // beats per minute
static int Division = 120;  // ticks per beat
static int NumTracks = 2;  // # tracks in MIDI file
static int Format = 1;  // MIDI format code

//
// parse midi data and put into streamitems of utterance
//
LISP midifile_to_utt(EST_String filename, LISP utt){

  int fd;
  short track;
  Event *event;
  Song mf;

  EST_Utterance *u = get_c_utt(utt);
  set_current_utt(u);  // keep ptr to current utt as static global
  
  if ((fd = open (filename.str(), O_RDONLY, 0)) == -1){
    cout << "couldn't open midifile " << filename  << endl;
    EST_error("...");
  } 
  mf.SMFRead(fd);   // read file
  
  // Track 0 configuration info - store as static globals
  Format    = mf.GetFormat();
  NumTracks = mf.GetNumTracks();
  Division  = mf.GetDivision();

  // while there are new MIDI events, read the event,
  //  and add it to the proper relation - uses functions below
  for (track=0; track<NumTracks; track++){
    for (event=mf.NextEvent(track); event!=NULL; event=mf.NextEvent(track)){
      Tclm_ParseEvent(event);
      //      Tclm_PrintEvent(cout, event);
      //      cout << endl;
    }
  }

  return utt;
}



void FL_add_lyric_token(unsigned long ltime, const char *lyrstr){
  // Add an item to the end of LyricToken relation
  // Also put in SylStructure relation and NoteStruct relation
  EST_Utterance *u = get_current_utt();  
  EST_Item *p;
  EST_String lyrString(lyrstr);

  if (lyrString.contains(RXalphanum)){ // skip non-alphanum 
    p = u->relation("LyricToken")->append();    
    p->set("name", lyrString);
    p->set("time", MIDIticks_to_sec(ltime));   
    u->relation("SylStructure")->append(p);
    u->relation("NoteStruct")->append(p);
  }
}


void FL_add_note(
		 unsigned long tstart,   // time start in ticks
		 int channel,            // MIDI channel (not used)
		 int pitch,              // pitch in range [0,127]
		 unsigned long duration, // duration in ticks
		 int velocity            // range [?,?]
		 ){

  EST_Utterance *u = get_current_utt();
  
  // Add a stream item to the end of Note stream
  // assume a monophonic track
  EST_Item *ni;
  float on,dur,freq;
  EST_String name;
  static float prev_off,prev_on;

  // if this is 1st note
  if (u->relation("Note")->length() == 0){
    prev_off = 0.0;
    prev_on  = -0.1;
  }

  ni = u->relation("Note")->append();
  MIDIpitch_to_freq_and_name(pitch, freq, name);

  // set features 
  ni->set("name",name); // musical note name

  on  = MIDIticks_to_sec(tstart);
  dur = MIDIticks_to_sec(duration);

  // if track is not unison, go back and fix previous note_off
  if (prev_off > on){
    prev(u->relation("Note")->tail())->set("off",on);
    
    if (prev_on >= on){  // simultaneous notes
      cout << "MIDI track is NOT unison - can't handle harmonies yet" << endl;
      EST_error("...");
    }
  }

  ni->set("on", on);
  ni->set("dur", dur);
  ni->set("off", on + dur);
  ni->set("freq",freq);
  ni->set("vel", velocity);
  ni->set("chan", channel);

  prev_off = on + dur;
}


void FL_add_pitchbend(unsigned long tstart,   // time start in ticks
		      int channel,            // MIDI channel (not used)
		      unsigned long value     // range [-8192,8192] (after correction)
		      ){
  // Add item to the end of F0_Mod stream
  //   - assumes a monophonic track
  EST_Item *item;
  float f_value, percent, semitones;
  EST_Utterance *u = get_current_utt();
  EST_Features *gl = get_Flinger_params();
  
  // figure out bend value
  f_value = ((float) value - 8192)/8192.0;  // center around 0 and normalize to [-1,1]
  semitones = gl->F("bend_semi_max") * f_value;
  percent = pow(2.0, semitones/12.0);
  //  printf("bend: MIDI=%d \t semi=%2.3f \t pcent=%1.2f\n",(int) value-8192,semitones,percent);
  
  // set features 
  item = u->relation("F0_Mod")->append();
  item->set("name", "pitchbend");
  item->set("semitones",semitones);            // bend value in semitones
  item->set("percent",percent);                // bend value in % of current freq
  item->set("start", MIDIticks_to_sec(tstart));
  item->set("chan", channel);                  // keep for future use?
}


void FL_add_vibrato(unsigned long tstart,   // time start in ticks
		    int channel,            // MIDI channel (not used)
		    unsigned long value     // range [0,127] 
		    ){

  // Add item to the end of F0_Mod stream
  // assume a monophonic track
  EST_Item *item;
  float percent;
  EST_Utterance *u = get_current_utt();
  EST_Features *gl = get_Flinger_params();
  
  // figure out vibrato value
  percent = gl->F("vibrato_max") * value/128.0; // normalize value to [0,1)
  
  // set features 
  item = u->relation("F0_Mod")->append();
  item->set("name", "vibrato");
  item->set("percent",percent);               // in % of current freq
  item->set("start", MIDIticks_to_sec(tstart));
  item->set("chan", channel);                 // keep for future use?
}



void FL_update_tempo(unsigned long time, short bpm){
  // It gets a little tricky if the tempo changes during the song
  //   - maintain an "anchor" from most recent tempo change
  PrevSecAnchor += Division * CurrentTempo * (time - PrevTicksAnchor);
  PrevTicksAnchor = time;
  CurrentTempo = bpm;

  cout << "update_tempo: " << PrevSecAnchor << " " <<  PrevTicksAnchor << " " << CurrentTempo << endl;
}

float MIDIticks_to_sec(int ticks){
  // Division = clock ticks per beat (I don't think this can vary in MIDI)
  // Tempo = beats per minute (This can vary during the song)  
  return (float) (ticks-PrevTicksAnchor)*60.0/(Division*CurrentTempo) + PrevSecAnchor;
}

void MIDIpitch_to_freq_and_name(int note_num, float &freq, EST_String &name){
  int octave = note_num / 12;
  octave -= 2;   // gives reasonable treble clef values
  EST_String s_octave;
  int scale_tone = note_num % 12;
  float base_freq;
  
  if ((scale_tone >= 0) && (scale_tone < 12)){
    base_freq = BaseFreq[scale_tone];
    freq = base_freq * (2 << octave);
    s_octave = (char) (48 + octave);
    name = NoteName[scale_tone] + s_octave;
  }
  else {
    cerr << "fl_midi_in:  scale tone=" << scale_tone << " out of range - how??\n.";
    EST_error("...");
  }
}

static EST_Utterance *get_current_utt(void){
  if (Utt != NULL)
    return Utt;
  else{
    EST_error("fl_midi_in:  utterance disappeared!\n");
    return Utt;
  }
}

static void set_current_utt(EST_Utterance *u){
  Utt = u;  return;
}
