#include <stdlib.h>
#include "festival.h"
#include "EST.h"
#include "fl_midi_in.h"
#include "fl_duration.h"
#include "fl_lyricize.h"
#include "fl_pitch.h"
#include "fl_token.h"
#include "fl_io.h"


LISP FL_sing_midifile(LISP filename);
static void sing_utt(LISP utt);
static LISP new_song_utt(void);
static LISP Flinger_Init(LISP params);


LISP FL_sing_midifile(LISP lfilename,LISP mode){
  cout << "* sing_midifile\n";
  LISP lutt=NIL;

  // if user has never called Flinger.init, set defaults
  if (!(Flinger_params_loaded()))
    Flinger_init_params(NIL);

  
  //
  // sing with a "mode" -- maybe we could use this later
  //  to say "sing with a jazzy feel"  etc...
  //
  LISP user_modes;
  user_modes = siod_get_lval("fling_modes",NULL);
  
  if ((mode == NIL) || (streq(get_c_string(mode),"default")))
    ;  // cout << "\tsing_midifile: default mode \n";
  else {
    cerr << "sing_midifile: can't find mode description \"" 
	 << get_c_string(mode) << "\" using default mode 'nil instead" << endl;
  }

  //
  // create empty utterance
  //
  lutt = new_song_utt();

  //
  // read midifile
  // parse midi data and put into items of utterance
  //
  EST_String filename = get_c_string(lfilename);
  midifile_to_utt(filename, lutt);

  //
  // Sing what has been encoded in the utterance struct
  //
  //  (when we get fancy, this will be inside a loop to enable
  //    pipelined output and multiple voices at a time)
  //
  sing_utt(lutt);
  
  return(lutt);
}


static void sing_utt(LISP utt){
  cout << "* sing_utt\n";
  // Sing this utt
  LISP lutt;

  // check whether utterance is empty
  if ((utt == NIL) ||
      (get_c_utt(utt)->relation("Note")->length() == 0))
    return;   // if so, do nothing
  
  lutt = quote(utt);
  
  // apply sing_hooks to utterance - synth and play it
  lutt = cons(rintern("apply_hooks"),
		cons(rintern("sing_hooks"),
		     cons(lutt,NIL)));
  lutt = cons(rintern("set!"),
	      cons(rintern("utt_sing"),
		   cons(lutt,NIL)));
  // sing it
  lutt = leval(lutt,NIL);
  user_gc(NIL); // this gives "bad env" error in lisp (?)
}


static LISP new_song_utt(void){

    // An empty utterance of type Song, ready to be filled from MIDI relation
    LISP lutt;

    // Use the LISP routine to do it
    lutt =  cons(rintern("Utterance"),
		 cons(rintern("Song"),
		      cons(NIL,NIL)));
    lutt = leval(lutt,NIL);
    EST_Utterance *u = get_c_utt(lutt);

    // linear relations
    u->create_relation("LyricToken");// like festival "Token"
    u->create_relation("Word");     // like festival "Word"
    u->create_relation("Syllable");
    u->create_relation("Segment");
    u->create_relation("Note");
    u->create_relation("F0_Mod"); // pitch bend wheel and vibrato (modulation)
    u->create_relation("Target"); // F0 contour -- I don't like the name, but 
                                  // it keeps us compatible with OGIresLPC 
                                  // for now.

    //
    // tree/multilinear relations across the above:
    //

    // LyricToken->Word->Syllable->Note->F0_Mod - used to get pitch contour, timing
    u->create_relation("NoteStruct"); 

    // LyricToken->Note - used to make Note, MIDI Lyric event association
    u->create_relation("NoteWordAssoc"); 

    // LyricToken->Word->Syllable->Segment - used to get pronunciations, timing
    u->create_relation("SylStructure");

    //
    // future
    //
    // u->create_relation("VQuality");  // (linear relation) voice quality, vocal effort
    // u->create_relation("DynStruct"); // Syllable->Note->Dyn_Mod
    //
    // u->create_relation("Feel");  // preciseness of hitting downbeats, note off's, etc.
    // u->create_relation("FeelStruct"); // Syllable->Note->Feel

    return lutt;
}


LISP FL_Initialize_Utt(LISP utt){
  cout << "* FL_Initialize_Utt\n";
    // Main utterance initialization
    EST_Utterance *u = get_c_utt(utt);
    EST_String type;
 
    type = utt_type(*u);
 
    utt_cleanup(*u);  // delete all relationss except IForm
 
    if (type == "Song"){
      ;  // nothing special to do yet...
    }
    else{
      // error
      cerr << "Unknown utterance type \"" << type << "\" for initialization "
	   << endl;
      EST_error("..");
    }
    
    return utt;
}
 


//
// Init called from Scheme.
//
static LISP Flinger_Init(LISP params){
  // extract LISP params, clear old ones first
  Flinger_init_params(params);
  
  return NIL;
}
LISP Flinger_AddInit(LISP params){
  // if user has never called Flinger.init, set defaults
  if (!(Flinger_params_loaded()))
    Flinger_init_params(NIL);

  // extract LISP params, don't clear old ones
  Flinger_addto_params(params);
  
  return NIL;
}


void festival_Flinger_init(void){
  //
  //  Initialization
  //
  cout << "* festival_Flinger_init \n"; 

  proclaim_module("Flinger");

  festival_def_utt_module("FL_Initialize",FL_Initialize_Utt,
			  "(FL_Initialize UTT)\n\
This module should be called first on all utterances.");

 festival_def_utt_module("FL_Tokenize",FL_Tokenize_Utt,
			  "(FL_Tokenize UTT)    \n\
Strips off (and keeps track of) punctuation symbols, etc.\n\
in LyricTokens.  Results in Word relation containing\n\
pronouncable words/syllables, no punctuation or symbols.");

 festival_def_utt_module("FL_Lyricize",FL_Lyricize_Utt,
			 "(FL_Lyricize UTT)    \n\
Figures out how to pronounce the lyrics.\n\
Turn the Word relation into Segments and Syllables: use\n\
- lexicon lookup\n\
- LtS rules\n\
- syllabification if needed\n\
- eventually fancier things like putting split\n\
words back together to lookup pronunciation,\n\\n\
then split again into syllables.\n\
Results in Segment relation with related Syllables.\n\
");

 festival_def_utt_module("FL_NoteDur",FL_NoteDur_Utt,
			 "(FL_NoteDur UTT)    \n\
Use relations between Segments, Syllables, Notes to get a \n\
duration for each Segment.  Also use default durations of\n\
Segments, other rules.   Result is duration for each Segment\n\
streamitem.\n\
");
 
 festival_def_utt_module("FL_NoteF0",FL_NoteF0_Utt,
			 "(FL_NoteF0 UTT)    \n\
Use relations between Segments, Notes, and ControlEvents to\n\
create an F0_Target relation.  This specifies target pitch, with\n\
attached features to describe vibrato, drift, etc.\n\
");
 
 init_subr_1("Flinger.init",Flinger_Init,
    "(Flinger_Init PARAMS)\n\
  Initialize parameters of vocalist (things that are not dynamically controllable). \n\
   PARAMS are an assoc list of parameter name and value.");
  
 init_subr_2("Flinger.sing",FL_sing_midifile,
	     "(Flinger.sing MIDIFILE MODE)\n\
Render MIDI data in MIDIFILE as singing.  Respect\n\
MODE.  Modes will be defined through the variable fling_modes (but not yet).");

 init_subr_2("builtin_lyrictoken_to_lyric",l_lyric_it,
	     "(builtin_lyrictoken_to_lyric LYRICTOKEN_RELATION LYRICTOKEN_NAME)\n\
Returns a list of words expanded from LYRICTOKEN_NAME.  Note that as this\n\
function may be called recursively. LYRICTOKEN_NAME might not be the name of\n\
LYRICTOKEN_RELATION depending on the rules applied.");
}


